"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.logErrAndExit = logErrAndExit;
exports.execAction = execAction;
exports.register = register;
exports.default = void 0;

function _defineProperty2() {
  const data = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

  _defineProperty2 = function () {
    return data;
  };

  return data;
}

function _bluebird() {
  const data = require("bluebird");

  _bluebird = function () {
    return data;
  };

  return data;
}

function _serializeError() {
  const data = require("serialize-error");

  _serializeError = function () {
    return data;
  };

  return data;
}

function _commander() {
  const data = _interopRequireDefault(require("commander"));

  _commander = function () {
    return data;
  };

  return data;
}

function _chalk() {
  const data = _interopRequireDefault(require("chalk"));

  _chalk = function () {
    return data;
  };

  return data;
}

function _didyoumean() {
  const data = _interopRequireDefault(require("didyoumean"));

  _didyoumean = function () {
    return data;
  };

  return data;
}

function _ink() {
  const data = require("ink");

  _ink = function () {
    return data;
  };

  return data;
}

function _consumer() {
  const data = require("../api/consumer");

  _consumer = function () {
    return data;
  };

  return data;
}

function _defaultErrorHandler() {
  const data = _interopRequireDefault(require("./default-error-handler"));

  _defaultErrorHandler = function () {
    return data;
  };

  return data;
}

function _utils() {
  const data = require("../utils");

  _utils = function () {
    return data;
  };

  return data;
}

function _loader() {
  const data = _interopRequireDefault(require("./loader"));

  _loader = function () {
    return data;
  };

  return data;
}

function _logger() {
  const data = _interopRequireDefault(require("../logger/logger"));

  _logger = function () {
    return data;
  };

  return data;
}

function _analytics() {
  const data = require("../analytics/analytics");

  _analytics = function () {
    return data;
  };

  return data;
}

function _constants() {
  const data = require("../constants");

  _constants = function () {
    return data;
  };

  return data;
}

function _globalFlags() {
  const data = _interopRequireDefault(require("./global-flags"));

  _globalFlags = function () {
    return data;
  };

  return data;
}

_didyoumean().default.returnFirstMatch = true;

function logErrAndExit(msg, commandName) {
  // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!
  if (msg.code) throw msg;
  console.error(msg); // eslint-disable-line

  _logger().default.exitAfterFlush(1, commandName);
}

function parseSubcommandFromArgs(args) {
  if (typeof (0, _utils().first)(args) === 'string') return (0, _utils().first)(args);
  return null;
}

function parseCommandName(commandName) {
  if (!commandName) return '';
  return (0, _utils().first)(commandName.split(' '));
}

function getOpts(c, opts) {
  const options = {};
  opts.forEach(([, name]) => {
    const parsedName = parseCommandName(name);
    const camelCaseName = (0, _utils().camelCase)(parsedName);

    if (name.startsWith('no-')) {
      // from commander help: "Note that multi-word options starting with --no prefix negate the boolean value of the following word. For example, --no-sauce sets the value of program.sauce to false."
      // we don't want this feature, so we do the opposite action.
      options[camelCaseName] = !c[(0, _utils().camelCase)(parsedName.replace('no-', ''))];
    } else {
      options[camelCaseName] = c[camelCaseName];
    }
  });
  return options;
}
/**
 * execute the command.
 * the stack trace up to this point is confusing, it's helpful to outline it here:
 * CLIExtension.run => commander.parse(params) => commander-pkg.parse => commander-pkg.parseArgs => execAction
 */


function execAction(command, concrete, args) {
  const flags = getOpts(concrete, command.options);
  const relevantArgs = args.slice(0, args.length - 1);
  const packageManagerArgs = concrete.parent.packageManagerArgs;

  _analytics().Analytics.init(concrete.name(), flags, relevantArgs, concrete.parent._version);

  _logger().default.info(`[*] started a new command: "${command.name}" with the following data:`, {
    args: relevantArgs,
    flags,
    packageManagerArgs
  });

  if (command.loader) {
    _loader().default.on();
  }

  if (flags[_constants().TOKEN_FLAG_NAME]) {
    _globalFlags().default.token = flags[_constants().TOKEN_FLAG_NAME].toString();
  }

  if (flags.json) {
    _loader().default.off();
  }

  _logger().default.shouldWriteToConsole = !flags.json;

  const migrateWrapper = run => {
    if (run) {
      _logger().default.debug('Checking if a migration is needed'); // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!


      return (0, _consumer().migrate)(null, false);
    }

    return Promise.resolve();
  };

  const getCommandHandler = () => {
    if (flags.json) {
      if (!command.json) throw new Error(`command "${command.name}" doesn't implement "json" method`);
      return 'json';
    }

    if (command.render && command.report) {
      return process.stdout.isTTY ? 'render' : 'report';
    }

    if (command.report) return 'report';
    if (command.render) return 'render';
    throw new Error(`command "${command.name}" doesn't implement "render" nor "report" methods`);
  };

  const commandHandler = getCommandHandler();
  return migrateWrapper(command.migration) // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!
  .then(() => {
    // this is a hack in the legacy code to make paper work
    // it should be removed upon major refactoring process
    command.packageManagerArgs = packageManagerArgs;
    return command[commandHandler](relevantArgs, flags, packageManagerArgs);
  }).then( /*#__PURE__*/function () {
    var _ref = (0, _bluebird().coroutine)(function* (res) {
      _loader().default.off();

      switch (commandHandler) {
        case 'json':
          {
            const code = res.code || 0;
            const data = res.data || res;
            return process.stdout.write(JSON.stringify(data, null, 2), () => _logger().default.exitAfterFlush(code, command.name));
          }

        case 'render':
          {
            const {
              waitUntilExit
            } = (0, _ink().render)(res);
            yield waitUntilExit();
            return _logger().default.exitAfterFlush(res.props.code, command.name);
          }

        case 'report':
        default:
          {
            return process.stdout.write(`${res.data}\n`, () => _logger().default.exitAfterFlush(res.code, command.name));
          }
      }
    });

    return function (_x) {
      return _ref.apply(this, arguments);
    };
  }()).catch(err => {
    _logger().default.error(`got an error from command ${command.name}: ${err}. Error serialized: ${JSON.stringify(err, Object.getOwnPropertyNames(err))}`);

    _loader().default.off();

    const errorHandled = (0, _defaultErrorHandler().default)(err) || command.handleError(err);
    if (command.private) return serializeErrAndExit(err, command.name); // uncomment this to see the entire error object on the console

    if (!command.private && errorHandled) return logErrAndExit(errorHandled, command.name);
    return logErrAndExit(err, command.name);
  });
}

function serializeErrAndExit(err, commandName) {
  process.stderr.write((0, _utils().packCommand)((0, _utils().buildCommandMessage)((0, _serializeError().serializeError)(err), undefined, false), false, false));
  const code = err.code && (0, _utils().isNumeric)(err.code) ? err.code : 1;
  return _logger().default.exitAfterFlush(code, commandName);
}
/**
 * register the action of each one of the commands.
 * at this point, it doesn't run any `execAction`, it only register it.
 * the actual running of `execAction` happens once `commander.parse(params)` is called.
 */


function registerAction(command, concrete) {
  concrete.action((...args) => {
    const subCommands = command.commands;

    if (subCommands === null || subCommands === void 0 ? void 0 : subCommands.length) {
      // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!
      const subcommandName = parseSubcommandFromArgs(args);
      const subcommand = subCommands.find(cmd => {
        return subcommandName === (parseCommandName(cmd.name) || cmd.alias);
      });
      args.shift();
      if (subcommand) return execAction(subcommand, concrete, args);
    }

    return execAction(command, concrete, args);
  });
}

function createOptStr(alias, name) {
  if (alias) {
    return `-${alias}, --${name}`;
  }

  return `--${name}`;
}

function register(command, commanderCmd) {
  const concrete = commanderCmd.command(command.name, null, {
    noHelp: command.private
  }).description(command.description).alias(command.alias);

  if (command.remoteOp) {
    (command.opts || command.options).push(['', _constants().TOKEN_FLAG, 'authentication token']);
  }

  (command.opts || command.options).forEach(([alias, name, description]) => {
    concrete.option(createOptStr(alias, name), description);
  }); // attach skip-update to all commands

  concrete.option(_constants().SKIP_UPDATE_FLAG, 'Skips auto updates');

  if (command.commands) {
    command.commands.forEach(nestedCmd => {
      register(nestedCmd, concrete);
    });
  }

  return registerAction(command, concrete);
}

class CommandRegistry {
  registerBaseCommand() {
    _commander().default.version(this.version).usage(this.usage).option(_constants().SKIP_UPDATE_FLAG, 'Skips auto updates for a command').description(this.description);
  }

  constructor(usage, description, version, commands, extensionsCommands) {
    (0, _defineProperty2().default)(this, "version", void 0);
    (0, _defineProperty2().default)(this, "usage", void 0);
    (0, _defineProperty2().default)(this, "description", void 0);
    (0, _defineProperty2().default)(this, "commands", void 0);
    (0, _defineProperty2().default)(this, "extensionsCommands", void 0);
    this.usage = usage;
    this.description = description;
    this.version = version;
    this.commands = commands; // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!

    this.extensionsCommands = extensionsCommands;
  }

  printHelp() {
    // eslint-disable-next-line global-require
    const helpTemplateGenerator = require('./templates/help');

    console.log(helpTemplateGenerator(this.extensionsCommands)); // eslint-disable-line no-console

    return this;
  }

  outputHelp() {
    const args = process.argv.slice(2);

    if (!args.length) {
      // @TODO replace back to commander help and override help method
      // commander.help();
      this.printHelp();
      return this;
    }

    const subcommand = args[0];
    const cmdList = this.commands.map(cmd => (0, _utils().first)(cmd.name.split(' '))); // @ts-ignore AUTO-ADDED-AFTER-MIGRATION-PLEASE-FIX!

    const extensionsCmdList = this.extensionsCommands.map(cmd => (0, _utils().first)(cmd.name.split(' ')));
    const aliasList = this.commands.map(cmd => (0, _utils().first)(cmd.alias.split(' ')));

    if (!cmdList.includes(subcommand) && !extensionsCmdList.includes(subcommand) && !aliasList.includes(subcommand) && subcommand !== '-V' && subcommand !== '--version') {
      process.stdout.write(_chalk().default.yellow(`warning: '${_chalk().default.bold(subcommand)}' is not a valid command.\nsee 'bit --help' for additional information.\n`));
      const suggestion = (0, _didyoumean().default)(subcommand, _commander().default.commands.filter(c => !c._noHelp).map(cmd => cmd._name));

      if (suggestion) {
        const match = typeof suggestion === 'string' ? suggestion : suggestion[0];
        console.log(_chalk().default.red(`Did you mean ${_chalk().default.bold(match)}?`)); // eslint-disable-line no-console
      }

      return this;
    }

    return this;
  }

}

exports.default = CommandRegistry;